/*
 * tcp_server.c
 *
 *
 */

#include <tcp_defs.h>
#include <tcp_server.h>
#include <lwip/tcp.h>
#include <lwip/stats.h>
#include <lwip/debug.h>
#include <lwip/pbuf.h>
#include <debug.h>


void tcp_server_receive(tcpPacket *msg) {
    struct tcp_server_struct *es;

    es = (struct tcp_server_struct *)msg->arg;

    struct pbuf *p = msg->p;

    /* if we receive an empty tcp frame from client => close connection */
    if (p == NULL) {
        /* remote host closed connection */
        es->state = ES_CLOSING;
        if (es->p == NULL) {
            /* we're done sending, close connection */
            debug_print("server receive: done sending, closing\n");
            tcp_server_close(msg->tpcb, es);
        }
        else {
            /* we're not done yet */
            /* acknowledge received packet */
            tcp_sent(msg->tpcb, tcp_server_sent);

            /* send remaining data*/
            debug_print("server receive: more to send\n");
            tcp_server_send(msg->tpcb, es);
        }
    }
    /* else : a non empty frame was received from client but for some reason err != ERR_OK */
    else if(msg->err != ERR_OK) {
        debug_print("server receive: err != ERR_OK\n");
        /* free received pbuf*/
        if (p != NULL) {
            es->p = NULL;
            pbuf_free(p);
        }
    }
    else if(es->state == ES_ACCEPTED) {
        /* first data chunk in p->payload */
        es->state = ES_RECEIVED;

        /* store reference to incoming pbuf (chain) */
        es->p = p;

        /* initialize LwIP tcp_sent callback function */
        tcp_sent(msg->tpcb, tcp_server_sent);

        /* send back the received data (echo) */
        debug_print("server receive: send first chunk\n");
        tcp_server_send(msg->tpcb, es);
    }
    else if (es->state == ES_RECEIVED) {
        /* more data received from client and previous data has been already sent*/
        if (es->p == NULL) {
            es->p = p;

            /* send back received data */
            tcp_server_send(msg->tpcb, es);
            debug_print("server receive: send next\n");
        }
        else {
            struct pbuf *ptr;

            /* chain pbufs to the end of what we recv'ed previously  */
            ptr = es->p;
            pbuf_chain(ptr,p);
        }
    }

    /* data received when connection already closed */
    else {
        /* Acknowledge data reception */
        tcp_recved(msg->tpcb, p->tot_len);

        /* free pbuf and do nothing */
        es->p = NULL;
        pbuf_free(p);
        debug_print("server receive: received but connection is closed\n");
    }

    tcp_output(msg->tpcb);
}

void tcp_server_send(struct tcp_pcb *tpcb, struct tcp_server_struct *es) {
    struct pbuf *ptr;
    err_t wr_err = ERR_OK;

    while ((wr_err == ERR_OK) &&
        (es->p != NULL) &&
        (es->p->len <= tcp_sndbuf(tpcb))) {

        /* get pointer on pbuf from es structure */
        ptr = es->p;

        /* enqueue data for transmission */
        wr_err = tcp_write(tpcb, ptr->payload, ptr->len, 1);
        debug_print("server send: enqueued\n");
        if (wr_err == ERR_OK) {
            u16_t plen;

            plen = ptr->len;

            /* continue with next pbuf in chain (if any) */
            es->p = ptr->next;

            if (es->p != NULL) {
                /* increment reference count for es->p */
                pbuf_ref(es->p);
            }

            /* free pbuf: will free pbufs up to es->p (because es->p has a reference count > 0) */
            pbuf_free(ptr);

            /* Update tcp window size to be advertized : should be called when received
            data (with the amount plen) has been processed by the application layer */
            tcp_recved(tpcb, plen);
        }
        else if(wr_err == ERR_MEM) {
            /* we are low on memory, try later / harder, defer to poll */
            es->p = ptr;
            debug_print("server send: err 1\n");
        }
        else {
            /* other problem ?? */
            debug_print("server send: err 2\n");
        }
    }
}

err_t tcp_server_sent(void *arg, struct tcp_pcb *tpcb, u16_t len) {
    struct tcp_server_struct *es;
    tcpPacket msg;
    es = (struct tcp_server_struct *)arg;

    if (es->p != NULL) {
        /* still got pbufs to send */
        debug_print("server sent: sending\n");
        tcp_server_send(tpcb, es);
    }
    else {
        /* if no more data to send and client closed connection*/
        if (es->state == ES_CLOSING) {
            debug_print("server sent: closing\n");
            tcp_server_close(tpcb, es);
        }
    }
    debug_print("server sent: done\n");
    return ERR_OK;
}

void tcp_server_errormsg(tcpPacket *msg) {
    struct tcp_server_struct *es;

    es = (struct tcp_server_struct *)msg->arg;
    if (es != NULL) {
        /*  free es structure */
        mem_free(es);
    }
    debug_print("server error: done\n");
}

void tcp_server_close(struct tcp_pcb *tpcb, struct tcp_server_struct *es) {
  /* remove all callbacks */
  tcp_arg(tpcb, NULL);
  tcp_sent(tpcb, NULL);
  tcp_recv(tpcb, NULL);
  tcp_err(tpcb, NULL);
  tcp_poll(tpcb, NULL, 0);

  /* delete es structure */
  if (es != NULL) {
    mem_free(es);
  }

  /* close tcp connection */
  tcp_close(tpcb);
  debug_print("server close: done\n");
}


